@@ -3,8 +3,10 @@
 # Distributed under the 3-clause BSD license, see accompanying file LICENSE
 # or https://github.com/scikit-hep/decaylanguage for details.
 
+from typing import Any, Dict, Iterator, List, Pattern, Tuple, Union
 
-def iter_flatten(iterable):
+
+def iter_flatten(iterable: Union[List[str], Tuple[str, ...]]) -> Iterator[str]:
     """
     Flatten nested tuples and lists
     """
@@ -15,7 +17,7 @@ def iter_flatten(iterable):
             yield e
 
 
-def split(x):
+def split(x: str) -> List[str]:
     """
     Break up a comma separated list, but respect curly brackets.
 
@@ -40,14 +42,16 @@ def split(x):
         elif x[i] == "}":
             c -= 1
         i += 1
+    return out
 
 
-def filter_lines(matcher, inp):
+def filter_lines(
+    matcher: Pattern[str], inp: List[str]
+) -> Tuple[List[Dict[str, Union[str, Any]]], List[str]]:
     """
     Filter out lines into new variable if they match a regular expression
     """
-    output = [
-        matcher.match(ln).groupdict() for ln in inp if matcher.match(ln) is not None
-    ]
+    matches = (matcher.match(ln) for ln in inp)
+    output = [match.groupdict() for match in matches if match is not None]
     new_inp = [ln for ln in inp if matcher.match(ln) is None]
     return output, new_inp
