@@ -25,6 +25,11 @@ def test_DaughtersDict_constructor_from_string():
     assert dd == {"K+": 1, "K-": 1, "pi0": 1}
 
 
+def test_DaughtersDict_constructor_kwargs():
+    dd = DaughtersDict("K+ K-", pi0=1, gamma=2)
+    assert dd == {"K+": 1, "K-": 1, "pi0": 1, "gamma": 2}
+
+
 def test_DaughtersDict_string_repr():
     dd = DaughtersDict(["K+", "K-", "K-", "pi+", "pi0"])
     assert dd.__str__() == "<DaughtersDict: ['K+', 'K-', 'K-', 'pi+', 'pi0']>"
@@ -47,6 +52,18 @@ def test_DaughtersDict_to_string():
     assert dd1.to_string() == "K+ K- K- pi0 pi0 pi0"
 
 
+def test_DaughtersDict_charge_conjugate():
+    dd = DaughtersDict({"K+": 2, "pi0": 1})
+    assert dd.charge_conjugate() == {"K-": 2, "pi0": 1}
+
+
+def test_DaughtersDict_charge_conjugate_pdg_names():
+    dd = DaughtersDict({"K(S)0": 1, "pi+": 1})  # PDG names!
+    assert dd == {"K(S)0": 1, "pi+": 1}  # PDG names kept as-is
+    assert dd.charge_conjugate() == {"ChargeConj(K(S)0)": 1, "pi-": 1}
+    assert dd.charge_conjugate(pdg_name=True) == {"K(S)0": 1, "pi-": 1}
+
+
 def test_DecayMode_constructor_default():
     dm = DecayMode()
     assert dm.bf == 0
@@ -285,7 +302,7 @@ def test_DecayChain_flatten_complex():
     assert dc_flatten.decays[dc_flatten.mother].daughters == DaughtersDict(
         {"gamma": 6, "pi+": 3, "pi-": 2}
     )
-    assert dc_flatten.bf == approx(1.0 * 0.0124 * (0.692 ** 2) * (0.98823 ** 3))
+    assert dc_flatten.bf == approx(1.0 * 0.0124 * (0.692**2) * (0.98823**3))
 
 
 def test_DecayChain_flatten_with_stable_particles():
@@ -300,7 +317,7 @@ def test_DecayChain_flatten_with_stable_particles():
     assert dc_flatten.decays[dc_flatten.mother].daughters == DaughtersDict(
         {"pi0": 4, "pi+": 3, "pi-": 2}
     )
-    assert dc_flatten.bf == approx(0.5 * (0.0124 ** 2) * (0.692 ** 2))
+    assert dc_flatten.bf == approx(0.5 * (0.0124**2) * (0.692**2))
 
 
 def test_DecayChain_string_repr(dc):
